from distutils.core import Command
from setuptools.command.test import test as TestCommand
from distutils.errors import DistutilsOptionError
from distutils import log
from subprocess import Popen, PIPE, call
import os
import sys
import time
import json


# Exception classes used by this module.
class CalledProcessError(Exception):
    """This exception is raised when a process run by check_call() or
    check_output() returns a non-zero exit status.
    The exit status will be stored in the returncode attribute;
    check_output() will also store the output in the output attribute.
    """
    def __init__(self, returncode, cmd, output=None):
        self.returncode = returncode
        self.cmd = cmd
        self.output = output

    def __str__(self):
        return "Command '%s' returned non-zero exit status %d" % (self.cmd,
                                                                  self
                                                                  .returncode)


def check_output(*popenargs, **kwargs):
    '''Run command with arguments and return its output as a byte string.
    If the exit code was non-zero it raises a CalledProcessError.  The
    CalledProcessError object will have the return code in the returncode
    attribute and output in the output attribute.
    The arguments are the same as for the Popen constructor.  Example:
    >>> check_output(["ls", "-l", "/dev/null"])
    'crw-rw-rw- 1 root root 1, 3 Oct 18  2007 /dev/null\n'
    The stdout argument is not allowed as it is used internally.
    To capture standard error in the result, use stderr=STDOUT.
    >>> import sys
    >>> check_output(["/bin/sh", "-c",
    ...               "ls -l non_existent_file ; exit 0"],
    ...              stderr=sys.stdout)
    'ls: non_existent_file: No such file or directory\n'
    '''
    if 'stdout' in kwargs:
        raise ValueError('stdout argument not allowed, it will be '
                         'overridden.')
    process = Popen(*popenargs, stdout=PIPE, **kwargs)
    output, unused_err = process.communicate()
    retcode = process.poll()
    if retcode:
        cmd = kwargs.get('args')
        if cmd is None:
            cmd = popenargs[0]
        raise CalledProcessError(retcode, cmd, output=output)
    return output


def get_node_ip(node='riak01'):
    state = json.loads(check_output(['docker', 'inspect', node]).decode())
    return state[0]['NetworkSettings']['IPAddress']


class docker(object):
    docker_submodule_path = 'aioriak/tests/docker'

    def use_docker(self):
        return bool(int(os.environ.get('DOCKER_CLUSTER', 0)))

    def _check_retcode(self, retcode, args):
        if retcode:
            cmd = args[0]
            raise CalledProcessError(retcode, cmd)
        else:
            return True

    def cluster_is_started(self, node='riak01'):
        return node in check_output(['docker', 'ps']).decode()


class docker_build(Command, docker):
    user_options = []
    description = 'Setup riak cluster with docker'
    verbose = False

    def initialize_options(self):
        self.verbose = bool(int(os.environ.get('VERBOSE', 0)))
        print(self.verbose)

    def finalize_options(self):
        pass

    def run(self):
        args = ['make', '-C', self.docker_submodule_path, 'build']
        retcode = call(args)
        self._check_retcode(retcode, args)

    sub_commands = [('docker_setup', lambda self: True)]


class docker_start(Command, docker):
    user_options = []
    description = 'Start riak cluster'
    verbose = False

    def initialize_options(self):
        self.verbose = bool(int(os.environ.get('VERBOSE', 0)))
        os.environ['DOCKER_RIAK_AUTOMATIC_CLUSTERING'] = '1'
        os.environ['DOCKER_RIAK_CLUSTER_SIZE'] = '3'
        os.environ['DOCKER_RIAK_BACKEND'] = 'memory'

    def finalize_options(self):
        pass

    def run(self):
        args = ['make', '-C', self.docker_submodule_path, 'start-cluster']
        retcode = call(args)
        if self._check_retcode(retcode, args):
            time.sleep(3)


class docker_stop(Command, docker):
    user_options = []
    description = 'Stop riak cluster'
    verbose = False

    def initialize_options(self):
        self.verbose = bool(int(os.environ.get('VERBOSE', 0)))

    def finalize_options(self):
        pass

    def run(self):
        if self.use_docker() and self.cluster_is_started():
            args = ['make', '-C', self.docker_submodule_path, 'stop-cluster']
            retcode = call(args)
            self._check_retcode(retcode, args)


class create_bucket_types(Command):
    '''
    Creates bucket-types appropriate for testing. By default this will create:
    * `pytest-maps` with ``{"datatype":"map"}``
    * `pytest-sets` with ``{"datatype":"set"}``
    * `pytest-counters` with ``{"datatype":"counter"}``
    * `pytest-consistent` with ``{"consistent":true}``
    * `pytest-write-once` with ``{"write_once": true}``
    * `pytest-mr`
    * `pytest` with ``{"allow_mult":false}``
    '''

    description = "create bucket-types used in integration tests"

    user_options = [
        ('riak-admin=', None, 'path to the riak-admin script')
    ]

    _props = {
        'pytest-maps': {'datatype': 'map'},
        'pytest-sets': {'datatype': 'set'},
        'pytest-counters': {'datatype': 'counter'},
        'pytest-consistent': {'consistent': True},
        'pytest-write-once': {'write_once': True},
        'pytest-mr': {},
        'pytest': {'allow_mult': False}
    }

    def initialize_options(self):
        self.riak_admin = None

    def finalize_options(self):
        if self.riak_admin is None:
            raise DistutilsOptionError("riak-admin option not set")

    def run(self):
        for name in self._props:
            self._create_and_activate_type(name, self._props[name])

    def check_btype_command(self, *args):
        cmd = self._btype_command(*args)
        return check_output(cmd)

    def _create_and_activate_type(self, name, props):
        # Check status of bucket-type
        exists = False
        active = False
        try:
            status = self.check_btype_command('status', name)
        except CalledProcessError as e:
            status = e.output

        exists = ('not an existing bucket type' not in status.decode('ascii'))
        active = ('is active' in status.decode('ascii'))

        if exists or active:
            log.info('Update {} bucket-type'.format(name))
            self.check_btype_command('update', name,
                                     json.dumps({'props': props},
                                                separators=(',', ':')))
        else:
            log.info('Create  {} bucket-type'.format(name))
            self.check_btype_command('create', name,
                                     json.dumps({'props': props},
                                                separators=(',', ':')))

        if not active:
            log.info('Activate {} bucket-type'.format(name))
            self.check_btype_command('activate', name)

    def _btype_command(self, *args):
        cmd = self.riak_admin + ['bucket-type']
        cmd.extend(args)
        return cmd


class setup_riak(Command, docker):
    user_options = []
    description = 'Setup riak cluster'
    verbose = False

    user_options = []

    def get_riak_admin(self):
        return os.environ.get('RIAK_ADMIN', None)

    def initialize_options(self):
        self.verbose = bool(int(os.environ.get('VERBOSE', 0)))
        self.riak_admin = self.get_riak_admin()

    def finalize_options(self):
        if self.riak_admin is None and not self.use_docker():
            raise DistutilsOptionError("riak-admin option not set")
        if self.use_docker():
            self.riak_admin = ['docker', 'exec', '-i', '-t', 'riak01',
                               'riak-admin']
        else:
            self.riak_admin = self.riak_admin.split()

    def run(self):
        bucket = self.distribution.get_command_obj('create_bucket_types')
        bucket.riak_admin = self.riak_admin

        for cmd_name in self.get_sub_commands():
            self.run_command(cmd_name)

    sub_commands = [
        ('docker_build', lambda self: self.use_docker()),
        ('docker_start',
         lambda self: self.use_docker() and not self.cluster_is_started()),
        ('create_bucket_types', None)]


class Test(TestCommand, docker):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import nose
        result = nose.run(argv=['nosetests'])
        if self.use_docker():
            self.run_command('docker_stop')
        sys.exit(not result)



from . import TYPES
from .datatype import Datatype


class Counter(Datatype):
    '''
    A convergent datatype that represents a counter which can be
    incremented or decremented. This type can stand on its own or be
    embedded within a :class:`~riak.datatypes.Map`.
    '''

    type_name = 'counter'
    _type_error_msg = 'Counters can only be integers'

    def _post_init(self):
        self._increment = 0

    def _default_value(self):
        return 0

    @Datatype.modified.getter
    def modified(self):
        '''
        Whether this counter has staged increments.
        '''
        return self._increment is not 0

    def to_op(self):
        '''
        Extracts the mutation operation from the counter
        :rtype: int, None
        '''
        if not self._increment == 0:
            return ('increment', self._increment)

    def increment(self, amount=1):
        '''
        Increments the counter by one or the given amount.

        :param amount: the amount to increment the counter
        :type amount: int
        '''
        self._raise_if_badtype(amount)
        self._increment += amount

    def decrement(self, amount=1):
        '''
        Decrements the counter by one or the given amount.

        :param amount: the amount to decrement the counter
        :type amount: int
        '''
        self._raise_if_badtype(amount)
        self._increment -= amount

    def _check_type(self, new_value):
        return isinstance(new_value, int)


TYPES['counter'] = Counter



from .datatype import Datatype
from . import TYPES
from collections import Mapping
from riak.util import lazy_property


class TypedMapView(Mapping):
    '''
    Implements a sort of view over a :class:`Map`, filtered by the embedded
    datatype.
    '''

    def __init__(self, parent, datatype):
        self.map = parent
        self.datatype = datatype

    # Mapping API
    def __getitem__(self, key):
        '''
        Fetches an item from the parent :class:`Map` scoped by this view's
        datatype.

        :param key: the key of the item
        :type key: str
        :rtype: :class:`~aioriak.datatypes.Datatype`
        '''
        return self.map[(key, self.datatype)]

    def __iter__(self):
        '''
        Iterates over all keys in the :class:`Map` scoped by this view's
        datatype.
        '''
        for key in self.map.value:
            name, datatype = key
            if datatype == self.datatype:
                yield name

    def __len__(self):
        '''
        Returns the number of keys in this map scoped by this view's datatype.
        '''
        return len(iter(self))

    def __contains__(self, key):
        '''
        Determines whether the given key with this view's datatype is in the
        parent :class:`Map`.
        '''
        return (key, self.datatype) in self.map

    # From the MutableMapping API
    def __delitem__(self, key):
        '''
        Removes the key with this view's datatype from the parent :class:`Map`.
        '''
        del self.map[(key, self.datatype)]


class Map(Mapping, Datatype):
    '''
    A convergent datatype that acts as a key-value datastructure. Keys
    are pairs of ``(name, datatype)`` where ``name`` is a string and
    ``datatype`` is the datatype name. Values are other convergent
    datatypes, represented by any concrete type in this module.

    You cannot set values in the map directly (it does not implement
    ``__setitem__``), but you may add new empty values or access
    non-existing values directly via bracket syntax. If a key is not in the
    original value of the map when accessed, fetching the key will cause
    its associated value to be created.::

        map[('name', 'register')]

    Keys and their associated values may be deleted from the map as
    you would in a dict::

        del map[('emails', 'set')]

    Convenience accessors exist that partition the map's keys by
    datatype and implement the :class:`~collections.Mapping`
    behavior as well as supporting deletion::

        map.sets['emails']
        map.registers['name']
        del map.counters['likes']
    '''

    type_name = 'map'
    _type_error_msg = "Map must be a dict with (name, type) keys"

    def _default_value(self):
        return dict()

    def _post_init(self):
        self._removes = set()
        self._updates = {}

    @lazy_property
    def counters(self):
        '''
        Filters keys in the map to only those of counter types.

        Example::
            map.counters['views'].increment()
            del map.counters['points']
        '''
        return TypedMapView(self, 'counter')

    @lazy_property
    def flags(self):
        '''
        Filters keys in the map to only those of flag types. Example::
            map.flags['confirmed'].enable()
            del map.flags['attending']
        '''
        return TypedMapView(self, 'flag')

    @lazy_property
    def maps(self):
        '''
        Filters keys in the map to only those of map types. Example::
            map.maps['emails'].registers['home'].set("user@example.com")
            del map.maps['spam']
        '''
        return TypedMapView(self, 'map')

    @lazy_property
    def registers(self):
        '''
        Filters keys in the map to only those of register types. Example::
            map.registers['username'].set_value("riak-user")
            del map.registers['access_key']
        '''
        return TypedMapView(self, 'register')

    @lazy_property
    def sets(self):
        '''
        Filters keys in the map to only those of set types. Example::
            map.sets['friends'].add("brett")
            del map.sets['favorites']
        '''
        return TypedMapView(self, 'set')

    def __contains__(self, key):
        '''
        A map contains a key if that key exists in the original value
        or has been added or mutated.
        :rtype: bool
        '''
        self._check_key(key)
        return (key in self._value) or (key in self._updates)

    # collections.Mapping API
    def __getitem__(self, key):
        '''
        Fetches a convergent datatype at the given key.

        .. note: If the key is not in the map, a new empty datatype
           will be inserted at that key and returned. If the key was
           previously deleted, that mutation will be discarded.

        :param key: the key of the value to fetch
        :type key: tuple
        :rtype: :class:`Datatype` matching the datatype in the key
        '''
        self._check_key(key)
        if key in self._value:
            return self._value[key]
        else:
            # If the key does not exist, we assume they are wanting to
            # create a new one with that name/type.
            if key not in self._updates:
                self._updates[key] = TYPES[key[1]](context=self.context)
            return self._updates[key]

    def __iter__(self):
        '''
        Iterates over the *immutable* original value of the map.
        '''
        return iter(self.value)

    def __len__(self):
        '''
        Returns the size of the original value of the map.
        '''
        return len(self._value)

    def __delitem__(self, key):
        '''
        Deletes a key from the map. If you have previously mutated the
        datatype associated with this key, those mutations will be
        discarded.

        .. note: You may delete keys that are not entries in the map.
           If the Riak server does not find the entry in the set, an
           error may be returned to the client. For safety, always
           submit removal operations with a context.

        :param key: the key to remove
        :type key: tuple
        '''
        # NB: deleting a key only marks it deleted, and you can delete
        # things that don't appear in the value!
        self._check_key(key)
        self._require_context()
        self._removes.add(key)

    def _check_key(self, key):
        '''
        Ensures well-formedness of a key.
        '''
        if not len(key) == 2:
            raise TypeError('invalid key: %r' % key)
        elif key[1] not in TYPES:
            raise TypeError('invalid datatype: %s' % key[1])

    # Datatype API
    @Datatype.value.getter
    def value(self):
        '''
        Returns a copy of the original map's value. Nested values are
        pure Python values as returned by :attr:`Datatype.value` from
        the nested types.

        :rtype: dict
        '''
        pvalue = {}
        for key in self._value:
            pvalue[key] = self._value[key].value
        return pvalue

    @Datatype.modified.getter
    def modified(self):
        '''
        Whether the map has staged local modifications.
        '''
        if self._removes:
            return True
        for v in self._value:
            if self._value[v].modified:
                return True
        for v in self._updates:
            if self._updates[v].modified:
                return True
        return False

    def to_op(self):
        '''
        Extracts the modification operation(s) from the map.

        :rtype: list, None
        '''
        removes = [('remove', r) for r in self._removes]
        value_updates = list(self._extract_updates(self._value))
        new_updates = list(self._extract_updates(self._updates))
        all_updates = removes + value_updates + new_updates
        if all_updates:
            return all_updates
        else:
            return None

    def _check_type(self, value):
        for key in value:
            try:
                self._check_key(key)
            except:
                return False
        return True

    def _coerce_value(self, new_value):
        cvalue = {}
        for key in new_value:
            cvalue[key] = TYPES[key[1]](value=new_value[key],
                                        context=self._context)
        return cvalue

    def _extract_updates(self, d):
        for key in d:
            if d[key].modified:
                yield ('update', key, d[key].to_op())


TYPES[Map.type_name] = Map



import collections
from aioriak.datatypes.datatype import Datatype
from aioriak.datatypes import TYPES


class Set(collections.Set, Datatype):
    '''A convergent datatype representing a Set with observed-remove
    semantics. Currently strings are the only supported value type.

    Example::
        myset.add('barista')
        myset.add('roaster')
        myset.add('brewer')

    Likewise they can simply be removed::
        myset.discard('barista')

    This datatype also implements the `Set ABC
    <https://docs.python.org/2/library/collections.html>`_, meaning it
    supports ``len()``, ``in``, and iteration.
    '''
    type_name = 'set'
    _type_error_msg = 'Sets can only be iterables of strings'

    def _post_init(self):
        self._adds = set()
        self._removes = set()

    def _default_value(self):
        return frozenset()

    def add(self, element):
        '''
        Adds an element to the set.

        .. note: You may add elements that already exist in the set.
           This may be used as an "assertion" that the element is a
           member.

        :param element: the element to add
        :type element: str
        '''
        self._check_element(element)
        self._adds.add(element)

    def discard(self, element):
        '''
        Removes an element from the set.

        .. note: You may remove elements from the set that are not
           present, but a context from the server is required.

        :param element: the element to remove
        :type element: str
        '''
        self._check_element(element)
        self._require_context()
        self._removes.add(element)

    def to_op(self):
        '''
        Extracts the modification operation from the set.

        :rtype: dict, None
        '''
        if not self._adds and not self._removes:
            return None
        changes = {}
        if self._adds:
            changes['adds'] = list(self._adds)
        if self._removes:
            changes['removes'] = list(self._removes)
        return changes

    @Datatype.modified.getter
    def modified(self):
        '''
        Whether this set has staged adds or removes.
        '''
        return len(self._removes | self._adds) > 0

    # collections.Set API, operates only on the immutable version
    def __contains__(self, element):
        return element in self.value

    def __iter__(self):
        return iter(self.value)

    def __len__(self):
        return len(self.value)

    def _coerce_value(self, new_value):
        return frozenset(new_value)

    def _check_type(self, new_value):
        if not isinstance(new_value, collections.Iterable):
            return False
        for element in new_value:
            if not isinstance(element, str):
                return False
        return True

    @classmethod
    def _check_element(cls, element):
        if not isinstance(element, str):
            raise TypeError(Set._type_error_msg)

TYPES[Set.type_name] = Set



from . import TYPES
from aioriak.error import ContextRequired


class Datatype:
    '''
    Base class for all convergent datatype wrappers. You will not use
    this class directly, but it does define some methods are common to
    all datatype wrappers.
    '''

    #: The string "name" of this datatype. Each datatype should set this.
    type_name = None

    def __init__(self, bucket=None, key=None, value=None, context=None):
        self.bucket = bucket
        self.key = key
        self._context = context
        if value is not None:
            self._set_value(value)
        else:
            self._set_value(self._default_value())
        self._post_init()

    @property
    def value(self):
        '''
        The pure, immutable value of this datatype, as a Python value,
        which is unique for each datatype.
        **NB**: Do not use this property to mutate data, as it will not
        have any effect. Use the methods of the individual type to effect
        changes. This value is guaranteed to be independent of any internal
        data representation.
        '''
        return self._value

    @property
    def context(self):
        '''
        The opaque context for this type, if it was previously fetched.

        :rtype: str
        '''
        if self._context:
            return self._context[:]

    @property
    def modified(self):
        '''
        Whether this datatype has staged local modifications.

        :rtype: bool
        '''
        raise NotImplementedError

    # Lifecycle methods

    async def reload(self, **params):
        '''
        Reloads the datatype from Riak.

        .. warning: This clears any local modifications you might have
           made.

        :rtype: :class:`Datatype`
        '''
        if not self.bucket:
            raise ValueError('bucket property not assigned')

        if not self.key:
            raise ValueError('key property not assigned')

        dtype, value, context = await self.bucket._client._fetch_datatype(
            self.bucket, self.key, **params)

        if not dtype == self.type_name:
            raise TypeError("Expected datatype {} but "
                            "got datatype {}".format(self.__class__,
                                                     TYPES[dtype]))

        self.clear()
        self._context = context
        self._set_value(value)
        return self

    async def delete(self, **params):
        '''
        Deletes the datatype from Riak. See :meth:`RiakClient.delete()
        <aioriak.client.RiakClient.delete>` for options.
        '''
        self.clear()
        self._context = None
        self._set_value(self._default_value())
        await self.bucket._client.delete(self, **params)
        return self

    async def update(self, **params):
        '''
        Sends locally staged mutations to Riak.

        :rtype: a subclass of :class:`~aioriak.datatypes.Datatype`
        '''
        if not self.modified:
            raise ValueError("No operation to perform")

        params.setdefault('return_body', True)
        await self.bucket._client.update_datatype(self, **params)
        self.clear()

        return self

    store = update

    def clear(self):
        '''
        Removes all locally staged mutations.
        '''
        self._post_init()

    def to_op(self):
        '''
        Extracts the mutation operation from this datatype, if any.
        Each type must implement this method, returning the
        appropriate operation, or `None` if there is no queued
        mutation.
        '''
        raise NotImplementedError

    # Private stuff

    def _set_value(self, value):
        self._raise_if_badtype(value)
        self._value = self._coerce_value(value)

    def _raise_if_badtype(self, new_value):
        if not self._check_type(new_value):
            raise TypeError(self._type_error_msg)

    def _check_type(self, new_value):
        '''
        Checks that initial values of the type are appropriate. Each
        type must implement this method.

        :rtype: bool
        '''
        raise NotImplementedError

    def _default_value(self):
        '''
        Returns what the initial value of an empty datatype should be.
        '''
        raise NotImplementedError

    def _coerce_value(self, new_value):
        '''
        Coerces the input value into the internal representation for
        the type. Datatypes may override this method.
        '''
        return new_value

    def _post_init(self):
        '''
        Called at the end of :meth:`__init__` so that subclasses can tweak
        their own setup without overriding the constructor.
        '''
        pass

    def _require_context(self):
        '''
        Raises an exception if the context is not present
        '''
        if not self._context:
            raise ContextRequired()



from aioriak.content import RiakContent
from aioriak.error import ConflictError


def content_property(name, doc=None):
    '''
    Delegates a property to the first sibling in a RiakObject, raising
    an error when the object is in conflict.
    '''
    def _setter(self, value):
        if len(self.siblings) == 0:
            # In this case, assume that what the user wants is to
            # create a new sibling inside an empty object.
            self.siblings = [RiakContent(self)]
        if len(self.siblings) != 1:
            raise ConflictError()
        setattr(self.siblings[0], name, value)

    def _getter(self):
        if len(self.siblings) == 0:
            return
        if len(self.siblings) != 1:
            raise ConflictError()
        return getattr(self.siblings[0], name)

    return property(_getter, _setter, doc=doc)


class RiakObject:
    '''
    The RiakObject holds meta information about a Riak object, plus the
    object's data.
    '''
    siblings = []

    def __init__(self, client, bucket, key=None):
        '''
        Construct a new RiakObject.

        :param client: A RiakClient object.
        :type client: :class:`RiakClient <aioriak.client.RiakClient>`
        :param bucket: A RiakBucket object.
        :type bucket: :class:`RiakBucket <aioriak.bucket.RiakBucket>`
        :param key: An optional key. If not specified, then the key
         is generated by the server when :func:`store` is called.
        :type key: string
        '''

        if key is not None and len(key) == 0:
            raise ValueError('Key name must either be "None"'
                             ' or a non-empty string.')
        self.client = client
        self.bucket = bucket
        self.key = key
        self.vclock = None
        self.siblings = [RiakContent(self)]
        self._resolver = None

    def _exists(self):
        if len(self.siblings) == 0:
            return False
        elif len(self.siblings) > 1:
            # Even if all of the siblings are tombstones, the object
            # essentially exists.
            return True
        else:
            return self.siblings[0].exists

    exists = property(_exists, None, doc='''
        Whether the object exists. This is only ``False`` when there
        are no siblings (the object was not found), or the solitary
        sibling is a tombstone.
        ''')

    async def reload(self):
        '''
        Reload the object from Riak. When this operation completes, the
        object could contain new metadata and a new value, if the object
        was updated in Riak since it was last retrieved.

        .. note:: Even if the key is not found in Riak, this will
           return a :class:`RiakObject`. Check the :attr:`exists`
           property to see if the key was found.

        :rtype: :class:`RiakObject`
        '''

        await self.client.get(self)
        return self

    async def store(self, return_body=True):
        '''
        Store the object in Riak. When this operation completes, the
        object could contain new metadata and possibly new data if Riak
        contains a newer version of the object according to the object's
        vector clock.

        :param return_body: if the newly stored object should be retrieved
        :type return_body: bool
        :rtype: :class:`RiakObject` '''
        if len(self.siblings) != 1:
            raise ConflictError("Attempting to store an invalid object, "
                                "resolve the siblings first")

        await self.client.put(self, return_body)

        return self

    async def delete(self):
        '''
        Delete this object from Riak.

        :rtype: :class:`RiakObject`
        '''
        await self.client.delete(self)
        self.clear()
        return self

    def clear(self):
        '''
        Reset this object.

        :rtype: RiakObject
        '''
        self.siblings = []
        return self

    data = content_property('data', doc='''
        The data stored in this object, as Python objects. For the raw
        data, use the `encoded_data` property. If unset, accessing
        this property will result in decoding the `encoded_data`
        property into Python values. The decoding is dependent on the
        `content_type` property and the bucket's registered decoders.
        ''')

    encoded_data = content_property('encoded_data', doc='''
        The raw data stored in this object, essentially the encoded
        form of the `data` property. If unset, accessing this property
        will result in encoding the `data` property into a string. The
        encoding is dependent on the `content_type` property and the
        bucket's registered encoders.
        ''')

    charset = content_property('charset', doc='''
        The character set of the encoded data as a string
        ''')

    content_encoding = content_property('content_encoding', doc='''
        The encoding (compression) of the encoded data. Valid values
        are identity, deflate, gzip
        ''')

    usermeta = content_property('usermeta', doc='''
        Arbitrary user-defined metadata dict, mapping strings to strings.
        ''')

    links = content_property('links', doc='''
        A set of bucket/key/tag 3-tuples representing links to other
        keys.
        ''')

    content_type = content_property('content_type', doc='''
        The MIME media type of the encoded data as a string
        ''')

    indexes = content_property('indexes', doc='''
        The set of secondary index entries, consisting of
        index-name/value tuples
        ''')

    def _get_resolver(self):
        if callable(self._resolver):
            return self._resolver
        elif self._resolver is None:
            return self.bucket.resolver
        else:
            raise TypeError("resolver is not a function")

    def _set_resolver(self, value):
        if value is None or callable(value):
            self._resolver = value
        else:
            raise TypeError("resolver is not a function")

    resolver = property(_get_resolver, _set_resolver,
                        doc='''The sibling-resolution function for this
                        object. If the resolver is not set, the
                        bucket's resolver will be used.''')



class RiakError(Exception):
    '''
    Base class for exceptions generated in the Riak API.
    '''
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class ConflictError(RiakError):
    '''
    Raised when an operation is attempted on a
    :class:`~aioriak.riak_object.RiakObject` that has more than one
    sibling.
    '''
    def __init__(self, message='Object in conflict'):
        super(ConflictError, self).__init__(message)


class ContextRequired(RiakError):
    '''
    This exception is raised when removals of map fields and set
    entries are attempted and the datatype hasn't been initialized
    with a context.
    '''

    _default_message = ('A context is required for remove operations, '
                        'fetch the datatype first')

    def __init__(self, message=None):
        super(ContextRequired, self).__init__(message or
                                              self._default_message)



from aioriak.datatypes import TYPES


class Bucket:
    '''
    The ``Bucket`` object allows you to access and change information
    about a Riak bucket, and provides async methods to create or retrieve
    objects within the bucket.
    '''

    def __init__(self, client, name, bucket_type):
        '''
        Returns a new ``Bucket`` instance.

        :param client: A :class:`RiakClient <aioriak.client.RiakClient>`
            instance
        :type client: :class:`RiakClient <aioriak.client.RiakClient>`
        :param name: The bucket name
        :type name: string
        :param bucket_type: The parent bucket type of this bucket
        :type bucket_type: :class:`BucketType`
        '''
        if not isinstance(name, str):
            raise TypeError('Bucket name must be a string')

        if not isinstance(bucket_type, BucketType):
            raise TypeError('Parent bucket type must be a BucketType instance')

        self._client = client
        self.name = name
        self.bucket_type = bucket_type
        self._encoders = {}
        self._decoders = {}
        self._resolver = None

    def _get_resolver(self):
        if callable(self._resolver):
            return self._resolver
        elif self._resolver is None:
            return self._client.resolver
        else:
            raise TypeError("resolver is not a function")

    def _set_resolver(self, value):
        if value is None or callable(value):
            self._resolver = value
        else:
            raise TypeError("resolver is not a function")

    resolver = property(_get_resolver, _set_resolver,
                        doc='''The sibling-resolution function for this
                        bucket. If the resolver is not set, the
                        client's resolver will be used.''')

    def get_decoder(self, content_type):
        '''
        Get the decoding function for the provided content type for
        this bucket.

        :param content_type: the requested media type
        :type content_type: str
        :rtype: function
        '''
        if content_type in self._decoders:
            return self._decoders[content_type]
        else:
            return self._client.get_decoder(content_type)

    def get_encoder(self, content_type):
        '''
        Get the encoding function for the provided content type for
        this bucket.

        :param content_type: the requested media type
        :type content_type: str
        :param content_type: Content type requested
        '''
        if content_type in self._encoders:
            return self._encoders[content_type]
        else:
            return self._client.get_encoder(content_type)

    def set_encoder(self, content_type, encoder):
        '''
        Set the encoding function for the provided content type for
        this bucket.

        :param content_type: the requested media type
        :type content_type: str
        :param encoder: an encoding function, takes a single object
            argument and returns a string data as single argument.
        :type encoder: function
        '''
        self._encoders[content_type] = encoder
        return self

    def set_decoder(self, content_type, decoder):
        '''
        Set the decoding function for the provided content type for
        this bucket.

        :param content_type: the requested media type
        :type content_type: str
        :param decoder: a decoding function, takes a string and
            returns a Python type
        :type decoder: function
        '''
        self._decoders[content_type] = decoder
        return self

    async def set_property(self, key, value):
        '''
        Set a bucket property.

        :param key: Property to set.
        :type key: string
        :param value: Property value.
        :type value: mixed
        '''
        return await self.set_properties({key: value})

    async def get_property(self, key):
        '''
        Retrieve a bucket property.

        :param key: The property to retrieve.
        :type key: string
        :rtype: mixed
        '''
        return (await self.get_properties())[key]

    async def set_properties(self, props):
        '''
        Set multiple bucket properties in one call.

        :param props: A dictionary of properties
        :type props: dict
        '''
        await self._client.set_bucket_props(self, props)

    async def get_properties(self):
        '''
        Retrieve a dict of all bucket properties.

        :rtype: dict
        '''
        return await self._client.get_bucket_props(self)

    async def get_keys(self):
        '''
        Return all keys within the bucket.

        :rtype: list of keys
        '''
        return await self._client.get_keys(self)

    async def get(self, key):
        '''
        Retrieve an :class:`~aioriak.riak_object.RiakObject` or
        :class:`~aioriak.datatypes.Datatype`, based on the presence and value
        of the :attr:`datatype <BucketType.datatype>` bucket property.

        :param key: Name of the key.
        :type key: string
        :rtype: :class:`RiakObject <aioriak.riak_object.RiakObject>` or
           :class:`~aioriak.datatypes.Datatype`
        '''
        if await self.bucket_type.get_datatype():
            return await self._client.fetch_datatype(self, key)
        from aioriak.riak_object import RiakObject
        obj = RiakObject(self._client, self, key)
        return await obj.reload()

    async def new(self, key=None, data=None, content_type='application/json',
                  encoded_data=None):
        '''
        A shortcut for manually instantiating a new
        :class:`~aioriak.riak_object.RiakObject` or a new
        :class:`~aioriak.datatypes.Datatype`, based on the presence and value
        of the :attr:`datatype <BucketType.datatype>` bucket property. When
        the bucket contains a :class:`~aioriak.datatypes.Datatype`, all
        arguments are ignored except ``key``, otherwise they are used to
        initialize the :class:`~aioriak.riak_object.RiakObject`.

        :param key: Name of the key. Leaving this to be None (default)
                    will make Riak generate the key on store.
        :type key: str
        :param data: The data to store in a
           :class:`~aioriak.riak_object.RiakObject`, see
           :attr:`RiakObject.data <aioriak.riak_object.RiakObject.data>`.
        :type data: object
        :param content_type: The media type of the data stored in the
           :class:`~aioriak.riak_object.RiakObject`, see
           :attr:`RiakObject.content_type
           <aioriak.riak_object.RiakObject.content_type>`.
        :type content_type: str
        :param encoded_data: The encoded data to store in a
           :class:`~aioriak.riak_object.RiakObject`, see
           :attr:`RiakObject.encoded_data
           <aioriak.riak_object.RiakObject.encoded_data>`.
        :type encoded_data: str
        :rtype: :class:`~aioriak.riak_object.RiakObject` or
                :class:`~aioriak.datatypes.Datatype`
        '''
        from aioriak import RiakObject
        datatype = await self.bucket_type.get_datatype()
        if datatype:
            return TYPES[datatype](bucket=self, key=key)

        obj = RiakObject(self._client, self, key)
        obj.content_type = content_type
        if data is not None:
            obj.data = data
        if encoded_data is not None:
            obj.encoded_data = encoded_data
        return obj

    async def delete(self, key, **kwargs):
        '''
        Deletes a key from Riak. Short hand for
        ``bucket.new(key).delete()``. See :meth:`RiakClient.delete()
        <riak.client.RiakClient.delete>` for options.

        :param key: The key for the object
        :type key: string
        :rtype: RiakObject
        '''
        return await (await self.new(key)).delete(**kwargs)

    def __repr__(self):
        if self.bucket_type.is_default():
            return '<RiakBucket {}>'.format(self.name)
        else:
            return '<RiakBucket {}/{}>'.format(self.bucket_type.name,
                                               self.name)


class BucketType:
    '''
    The ``BucketType`` object allows you to access and change
    properties on a Riak bucket type and access buckets within its
    namespace.

    Async implementation of <riak.bucket.BucketType>
    '''
    def __init__(self, client, name):
        '''
        Returns a new ``BucketType`` instance.

        :param client: A :class:`RiakClient <aioriak.client.RiakClient>`
            instance
        :type client: :class:`RiakClient <aioriak.client.RiakClient>`
        :param name: The bucket-type's name
        :type name: string
        '''
        self._client = client
        self.name = name

    def __repr__(self):
        return "<BucketType {0}>".format(self.name)

    def __hash__(self):
        return hash((self.name, self._client))

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return hash(self) == hash(other)
        else:
            return False

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return hash(self) != hash(other)
        else:
            return True

    def _get_resolver(self):
        if callable(self._resolver):
            return self._resolver
        elif self._resolver is None:
            return self._client.resolver
        else:
            raise TypeError("resolver is not a function")

    def _set_resolver(self, value):
        if value is None or callable(value):
            self._resolver = value
        else:
            raise TypeError("resolver is not a function")

    resolver = property(_get_resolver, _set_resolver,
                        doc='''The sibling-resolution function for this
                        bucket. If the resolver is not set, the
                        client's resolver will be used.''')

    def is_default(self):
        '''
        Whether this bucket type is the default type, or a user-defined type.

        :rtype: bool
        '''
        return self.name == 'default'

    async def get_property(self, key):
        '''
        Retrieve a bucket-type property.

        :param key: The property to retrieve.
        :type key: string
        :rtype: mixed
        '''
        return await self.get_properties()[key]

    async def set_property(self, key, value):
        '''
        Set a bucket-type property.

        :param key: Property to set.
        :type key: string
        :param value: Property value.
        :type value: mixed
        '''
        await self.set_properties({key: value})

    async def get_properties(self):
        '''
        Retrieve a dict of all bucket-type properties.

        :rtype: dict
        '''
        return await self._client.get_bucket_type_props(self)

    async def set_properties(self, props):
        '''
        Set multiple bucket-type properties in one call.

        :param props: A dictionary of properties
        :type props: dict
        '''
        await self._client.set_bucket_type_props(self, props)

    def bucket(self, name):
        '''
        Gets a bucket that belongs to this bucket-type.

        :param name: the bucket name
        :type name: str
        :rtype: :class:`RiakBucket`
        '''
        return self._client.bucket(name, self)

    async def get_buckets(self):
        '''
        Get the list of buckets under this bucket-type as
        :class:`Bucket <aioriak.bucket.Bucket>` instances.

        .. warning:: Do not use this in production, as it requires
           traversing through all keys stored in a cluster.

        :rtype: list of :class:`Bucket <riak.bucket.Bucket>`
                instances
        '''
        return await self._client.get_buckets(bucket_type=self)

    async def get_datatype(self):
        '''
        The assigned datatype for this bucket type, if present.

        :rtype: None or string
        '''
        if not hasattr(self, '_datatype'):
            self._datatype = (await self.get_properties()).get('datatype')
        return self._datatype



import logging
import asyncio
import struct
import riak_pb
from riak_pb import messages
from riak.transports.pbc import codec
from aioriak.content import RiakContent
from riak.riak_object import VClock
from riak.util import decode_index_value, bytes_to_str, str_to_bytes
from aioriak.error import RiakError


MAX_CHUNK_SIZE = 65536

logger = logging.getLogger('aioriak.transport')


async def create_transport(host='localhost', port=8087, loop=None):
    reader, writer = await asyncio.open_connection(
        host, port, loop=loop)
    conn = RiakPbcAsyncTransport(reader, writer, loop=loop)
    return conn


class RPBPacketParser:
    ''' Riak protobuf packet parser.'''
    HEADER_LENGTH = 4

    def __init__(self, reader, initial_data=bytearray(), loop=None):
        self._loop = loop or asyncio.get_event_loop()
        self._data = initial_data
        self._reader = reader
        self._header_parsed = False
        self._tail = bytearray()
        self._eof = False

    @property
    def tail(self):
        return self._tail

    def _parse_header(self):
        ''' Parse protobuf message header '''
        if self._header_parsed:
            return True
        if len(self._data) >= self.HEADER_LENGTH:
            self._msglen, = struct.unpack(
                '!i', self._data[:self.HEADER_LENGTH])
            if self._msglen > 8192:
                raise RiakError('Wrong MESSAGE_LEN %d' % self._msglen)
            self._header_parsed = True
        else:
            self._header_parsed = False
        return self._header_parsed

    def _parse_msg(self):
        ''' Parse protobuf message'''
        self._msg = self._data[
            self.HEADER_LENGTH:self.HEADER_LENGTH + self._msglen]
        self.msg_code, = struct.unpack("B", self._msg[:1])
        if self.msg_code is messages.MSG_CODE_ERROR_RESP:
            error = self._get_pb_msg(self.msg_code, self._msg[1:])
            logger.error('Riak error message recieved: %s',
                         bytes_to_str(error.errmsg))
            raise RiakError(bytes_to_str(error.errmsg))
        elif self.msg_code in messages.MESSAGE_CLASSES:
            logger.debug('Normal message with code %d received', self.msg_code)
            self.msg = self._get_pb_msg(self.msg_code, self._msg[1:])
        else:
            logger.error('Unknown message received [%d]', self.msg_code)

    def _grow_tail(self):
        if len(self._data) > self._msglen + self.HEADER_LENGTH:
            self._tail = self._data[
                self.HEADER_LENGTH + self._msglen:]
        else:
            self._tail = bytearray()

    def _check_eof(self):
        if self._header_parsed and \
                len(self._data) >= self.HEADER_LENGTH + self._msglen:
            self._eof = True
        return self._eof

    def _get_pb_msg(self, code, msg):
        ''' Return protobuf object from raw mesage'''
        try:
            pbclass = messages.MESSAGE_CLASSES[code]
        except KeyError:
            pbclass = None

        if pbclass is None:
            return None
        pbo = pbclass()
        pbo.ParseFromString(bytes(msg))
        return pbo

    def at_eof(self):
        return self._eof

    async def get_pbo(self):
        if self._parse_header():
            if self._check_eof():
                self._parse_msg()
                self._grow_tail()
                return self.msg_code, self.msg

        while not self.at_eof():
            chunk = await self._reader.read(MAX_CHUNK_SIZE)
            self._data.extend(chunk)
            self._parse_header()
            if self._check_eof():
                self._parse_msg()
                self._grow_tail()
                return self.msg_code, self.msg


class RPBStreamParser:
    '''
    Riak protobuf stream packets parser
    This class is async generator with feed_data method
    and iterable packets on stream
    '''

    def __init__(self, reader, loop=None):
        self._loop = loop or asyncio.get_event_loop()
        self._reader = reader
        self._in_buf = bytearray()
        self.finished = False

    @property
    def tail(self):
        return self._in_buf

    async def __aiter__(self):
        return self

    async def __anext__(self):
        if self.finished:
            raise StopAsyncIteration
        msg_code, pbo = await self._fetch_pbo()
        if msg_code is not None:
            if pbo.done:
                self.finished = True
            return (msg_code, pbo)
        else:
            raise StopAsyncIteration

    async def _fetch_pbo(self):
        parser = RPBPacketParser(self._reader, self._in_buf, self._loop)
        code, pbo = await parser.get_pbo()
        self._in_buf = parser.tail
        return code, pbo


class RiakPbcAsyncTransport:
    ParserClass = RPBPacketParser
    StreamParserClass = RPBStreamParser

    def __init__(self, reader, writer, loop=None):
        self._loop = loop or asyncio.get_event_loop()
        self._writer = writer
        self._reader = reader
        self._parser = None

    def _encode_content(self, robj, rpb_content):
        '''
        Fills an RpbContent message with the appropriate data and
        metadata from a RiakObject.

        :param robj: a RiakObject
        :type robj: RiakObject
        :param rpb_content: the protobuf message to fill
        :type rpb_content: riak_pb.RpbContent
        '''
        if robj.content_type:
            rpb_content.content_type = robj.content_type.encode()
        if robj.charset:
            rpb_content.charset = robj.charset.encode()
        if robj.content_encoding:
            rpb_content.content_encoding = robj.content_encoding
        for uk in robj.usermeta:
            pair = rpb_content.usermeta.add()
            pair.key = uk.encode()
            pair.value = robj.usermeta[uk].encode()
        for link in robj.links:
            pb_link = rpb_content.links.add()
            try:
                bucket, key, tag = link
            except ValueError:
                raise RiakError("Invalid link tuple %s" % link)

            pb_link.bucket = bucket
            pb_link.key = key
            if tag:
                pb_link.tag = tag
            else:
                pb_link.tag = ''

        for field, value in robj.indexes:
            pair = rpb_content.indexes.add()
            pair.key = field
            pair.value = value.encode()

        rpb_content.value = robj.encoded_data

    def _encode_bucket_props(self, props, msg):
        '''
        Encodes a dict of bucket properties into the protobuf message.

        :param props: bucket properties
        :type props: dict
        :param msg: the protobuf message to fill
        :type msg: riak_pb.RpbSetBucketReq
        '''
        for prop in codec.NORMAL_PROPS:
            if prop in props and props[prop] is not None:
                if isinstance(props[prop], str):
                    prop_value = props[prop].encode()
                else:
                    prop_value = props[prop]
                setattr(msg.props, prop, prop_value)
        for prop in codec.COMMIT_HOOK_PROPS:
            if prop in props:
                setattr(msg.props, 'has_' + prop, True)
                self._encode_hooklist(props[prop], getattr(msg.props, prop))
        for prop in codec.MODFUN_PROPS:
            if prop in props and props[prop] is not None:
                self._encode_modfun(props[prop], getattr(msg.props, prop))
        for prop in codec.QUORUM_PROPS:
            if prop in props and props[prop] not in (None, 'default'):
                value = self._encode_quorum(props[prop])
                if value is not None:
                    setattr(msg.props, prop, value)
        if 'repl' in props:
            msg.props.repl = codec.REPL_TO_PY[props['repl']]

        return msg

    def _encode_hooklist(self, hooklist, msg):
        '''
        Encodes a list of commit hooks into their protobuf equivalent.
        Used in bucket properties.

        :param hooklist: a list of commit hooks
        :type hooklist: list
        :param msg: a protobuf field that is a list of commit hooks
        '''
        for hook in hooklist:
            pbhook = msg.add()
            self._encode_hook(hook, pbhook)

    def _encode_hook(self, hook, msg):
        '''
        Encodes a commit hook dict into the protobuf message. Used in
        bucket properties.

        :param hook: the hook to encode
        :type hook: dict
        :param msg: the protobuf message to fill
        :type msg: riak_pb.RpbCommitHook
        :rtype riak_pb.RpbCommitHook
        '''
        if 'name' in hook:
            msg.name = hook['name']
        else:
            self._encode_modfun(hook, msg.modfun)
        return msg

    def _encode_modfun(self, props, msg=None):
        '''
        Encodes a dict with 'mod' and 'fun' keys into a protobuf
        modfun pair. Used in bucket properties.

        :param props: the module/function pair
        :type props: dict
        :param msg: the protobuf message to fill
        :type msg: riak_pb.RpbModFun
        :rtype riak_pb.RpbModFun
        '''
        if msg is None:
            msg = riak_pb.RpbModFun()
        msg.module = props['mod'].encode()
        msg.function = props['fun'].encode()
        return msg

    def _encode_quorum(self, rw):
        '''
        Converts a symbolic quorum value into its on-the-wire
        equivalent.

        :param rw: the quorum
        :type rw: string, integer
        :rtype: integer
        '''
        if rw in codec.QUORUM_TO_PB:
            return codec.QUORUM_TO_PB[rw]
        elif type(rw) is int and rw >= 0:
            return rw
        else:
            return None

    def _encode_message(self, msg_code, msg=None):
        if msg is None:
            return struct.pack("!iB", 1, msg_code)
        msgstr = msg.SerializeToString()
        slen = len(msgstr)
        hdr = struct.pack("!iB", 1 + slen, msg_code)
        return hdr + msgstr

    @classmethod
    def _decode_pbo(cls, message):
        result = {}
        for key, value in message.ListFields():
            result[key.name] = value
        return result

    def _encode_dt_op(self, dtype, req, op):
        if dtype == 'counter':
            req.op.counter_op.increment = op[1]
        elif dtype == 'set':
            self._encode_set_op(req.op, op)
        elif dtype == 'map':
            self._encode_map_op(req.op.map_op, op)
        else:
            raise TypeError("Cannot send operation on datatype {!r}".
                            format(dtype))

    def _encode_set_op(self, msg, op):
        if 'adds' in op:
            msg.set_op.adds.extend(str_to_bytes(op['adds']))
        if 'removes' in op:
            msg.set_op.removes.extend(str_to_bytes(op['removes']))

    def _encode_map_op(self, msg, ops):
        for op in ops:
            name, dtype = op[1]
            ftype = codec.MAP_FIELD_TYPES[dtype]
            if op[0] == 'add':
                add = msg.adds.add()
                add.name = str_to_bytes(name)
                add.type = ftype
            elif op[0] == 'remove':
                remove = msg.removes.add()
                remove.name = str_to_bytes(name)
                remove.type = ftype
            elif op[0] == 'update':
                update = msg.updates.add()
                update.field.name = str_to_bytes(name)
                update.field.type = ftype
                self._encode_map_update(dtype, update, op[2])

    def _encode_map_update(self, dtype, msg, op):
        if dtype == 'counter':
            # ('increment', some_int)
            msg.counter_op.increment = op[1]
        elif dtype == 'set':
            self._encode_set_op(msg, op)
        elif dtype == 'map':
            self._encode_map_op(msg.map_op, op)
        elif dtype == 'register':
            # ('assign', some_str)
            msg.register_op = str_to_bytes(op[1])
        elif dtype == 'flag':
            if op == 'enable':
                msg.flag_op = riak_pb.MapUpdate.ENABLE
            else:
                msg.flag_op = riak_pb.MapUpdate.DISABLE

    def _encode_dt_options(self, req, params):
        for q in ['r', 'pr', 'w', 'dw', 'pw']:
            if q in params and params[q] is not None:
                setattr(req, q, self._encode_quorum(params[q]))

        for o in ['basic_quorum', 'notfound_ok', 'timeout', 'return_body',
                  'include_context']:
            if o in params and params[o] is not None:
                setattr(req, o, params[o])

    def _decode_dt_fetch(self, resp):
        dtype = codec.DT_FETCH_TYPES.get(resp.type)
        if dtype is None:
            raise ValueError("Unknown datatype on wire: {}".format(resp.type))

        value = self._decode_dt_value(dtype, resp.value)

        if resp.HasField('context'):
            context = resp.context[:]
        else:
            context = None

        return dtype, value, context

    def _decode_dt_value(self, dtype, msg):
        if dtype == 'counter':
            return msg.counter_value
        elif dtype == 'set':
            return self._decode_set_value(msg.set_value)
        elif dtype == 'map':
            return self._decode_map_value(msg.map_value)

    def _decode_map_value(self, entries):
        out = {}
        for entry in entries:
            name = entry.field.name[:].decode()
            dtype = codec.MAP_FIELD_TYPES[entry.field.type]
            if dtype == 'counter':
                value = entry.counter_value
            elif dtype == 'set':
                value = self._decode_set_value(entry.set_value)
            elif dtype == 'register':
                value = entry.register_value[:].decode()
            elif dtype == 'flag':
                value = entry.flag_value
            elif dtype == 'map':
                value = self._decode_map_value(entry.map_value)
            out[(name, dtype)] = value
        return out

    def _decode_set_value(self, set_value):
        return [string[:].decode() for string in set_value]

    def _decode_bucket_props(self, msg):
        '''
        Decodes the protobuf bucket properties message into a dict.

        :param msg: the protobuf message to decode
        :type msg: riak_pb.RpbBucketProps
        :rtype dict
        '''
        props = {}

        for prop_name in codec.NORMAL_PROPS:
            if msg.HasField(prop_name):
                prop = getattr(msg, prop_name)
                if isinstance(prop, bytes):
                    props[prop_name] = prop.decode()
                else:
                    props[prop_name] = prop
        for prop in codec.COMMIT_HOOK_PROPS:
            if getattr(msg, 'has_' + prop):
                props[prop] = self._decode_hooklist(getattr(msg, prop))
        for prop in codec.MODFUN_PROPS:
            if msg.HasField(prop):
                props[prop] = self._decode_modfun(getattr(msg, prop))
        for prop in codec.QUORUM_PROPS:
            if msg.HasField(prop):
                props[prop] = self._decode_quorum(getattr(msg, prop))
        if msg.HasField('repl'):
            props['repl'] = codec.REPL_TO_PY[msg.repl]

        return props

    def _decode_hooklist(self, hooklist):
        '''
        Decodes a list of protobuf commit hooks into their python
        equivalents. Used in bucket properties.

        :param hooklist: a list of protobuf commit hooks
        :type hooklist: list
        :rtype list
        '''
        return [self._decode_hook(hook) for hook in hooklist]

    def _decode_hook(self, hook):
        '''
        Decodes a protobuf commit hook message into a dict. Used in
        bucket properties.

        :param hook: the hook to decode
        :type hook: riak_pb.RpbCommitHook
        :rtype dict
        '''
        if hook.HasField('modfun'):
            return self._decode_modfun(hook.modfun)
        else:
            return {'name': hook.name}

    def _decode_modfun(self, modfun):
        '''
        Decodes a protobuf modfun pair into a dict with 'mod' and
        'fun' keys. Used in bucket properties.

        :param modfun: the protobuf message to decode
        :type modfun: riak_pb.RpbModFun
        :rtype dict
        '''
        return {'mod': modfun.module.decode(),
                'fun': modfun.function.decode()}

    def _decode_quorum(self, rw):
        '''
        Converts a protobuf quorum value to a symbolic value if
        necessary.

        :param rw: the quorum
        :type rw: int
        :rtype int or string
        '''
        if rw in codec.QUORUM_TO_PY:
            return codec.QUORUM_TO_PY[rw]
        else:
            return rw

    async def _stream(self, msg_code, msg=None, expect=None):
        self._writer.write(self._encode_message(msg_code, msg))
        self._parser = self.StreamParserClass(self._reader)
        responses = []
        async for code, pbo in self._parser:
            if expect is not None and code != expect:
                raise Exception('Unexpected response code ({})'.format(code))
            responses.append((code, pbo))
        return responses

    async def _request(self, msg_code, msg=None, expect=None):
        self._writer.write(self._encode_message(msg_code, msg))
        if self._parser:
            tail = self._parser.tail
            del self._parser
        else:
            tail = bytearray()
        self._parser = self.ParserClass(self._reader, tail)
        code, response = await self._parser.get_pbo()

        if expect is not None and code != expect:
            raise Exception('Unexpected response code ({})'.format(code))
        return code, response

    async def _read_response(self):
        while not self._reader.at_eof():
            try:
                data = await self._reader.read(MAX_CHUNK_SIZE)
            except asyncio.CancelledError:
                break
            except Exception:
                # XXX: for QUIT command connection error can be received
                #       before response
                # logger.error("Exception on data read %r", exc, exc_info=True)
                break
            self._parser.feed_data(data)
            if self._parser.at_eof():
                break
        return self._parser.msg_code, self._parser.msg

    def close(self):
        self._writer.close()

    async def ping(self):
        code, res = await self._request(messages.MSG_CODE_PING_REQ)
        if code == messages.MSG_CODE_PING_RESP:
            return True
        return False

    async def get_server_info(self):
        _, res = await self._request(
            messages.MSG_CODE_GET_SERVER_INFO_REQ,
            expect=messages.MSG_CODE_GET_SERVER_INFO_RESP)
        return self._decode_pbo(res)

    async def get_client_id(self):
        _, res = await self._request(
            messages.MSG_CODE_GET_CLIENT_ID_REQ,
            expect=messages.MSG_CODE_GET_CLIENT_ID_RESP)
        return self._decode_pbo(res)['client_id']

    async def set_client_id(self, client_id):
        req = riak_pb.RpbSetClientIdReq()
        req.client_id = client_id

        code, res = await self._request(
            messages.MSG_CODE_SET_CLIENT_ID_REQ, req,
            expect=messages.MSG_CODE_SET_CLIENT_ID_RESP)
        if code == messages.MSG_CODE_SET_CLIENT_ID_RESP:
            return True
        else:
            return False

    async def get_bucket_type_props(self, bucket_type):
        '''
        Fetch bucket-type properties

        :param bucket_type: A :class:`BucketType <aioriak.bucket.BucketType>`
               instance
        :type bucket_type: :class:`BucketType <aioriak.bucket.BucketType>`
        '''
        req = riak_pb.RpbGetBucketTypeReq()
        req.type = bucket_type.name.encode()

        msg_code, resp = await self._request(
            messages.MSG_CODE_GET_BUCKET_TYPE_REQ, req,
            messages.MSG_CODE_GET_BUCKET_RESP)
        return self._decode_bucket_props(resp.props)

    async def fetch_datatype(self, bucket, key):

        if bucket.bucket_type.is_default():
            raise NotImplementedError('Datatypes cannot be used in the default'
                                      ' bucket-type.')
        req = riak_pb.DtFetchReq()
        req.type = bucket.bucket_type.name.encode()
        req.bucket = bucket.name.encode()
        req.key = key.encode()

        msg_code, resp = await self._request(messages.MSG_CODE_DT_FETCH_REQ,
                                             req,
                                             messages.MSG_CODE_DT_FETCH_RESP)

        return self._decode_dt_fetch(resp)

    async def get_bucket_props(self, bucket):
        '''
        Serialize bucket property request and deserialize response
        '''
        req = riak_pb.RpbGetBucketReq()
        req.bucket = str_to_bytes(bucket.name)
        self._add_bucket_type(req, bucket.bucket_type)

        msg_code, resp = await self._request(
            messages.MSG_CODE_GET_BUCKET_REQ, req,
            messages.MSG_CODE_GET_BUCKET_RESP)

        return self._decode_bucket_props(resp.props)

    async def set_bucket_type_props(self, bucket_type, props):
        '''
        Set bucket-type properties

        :param bucket_type: A :class:`BucketType <aioriak.bucket.BucketType>`
               instance
        :type bucket_type: :class:`BucketType <aioriak.bucket.BucketType>`
        '''
        req = riak_pb.RpbSetBucketTypeReq()
        req.type = bucket_type.name.encode()

        self._encode_bucket_props(props, req)

        msg_code, resp = await self._request(
            messages.MSG_CODE_SET_BUCKET_TYPE_REQ, req,
            messages.MSG_CODE_SET_BUCKET_RESP)
        return True

    async def set_bucket_props(self, bucket, props):
        '''
        Serialize set bucket property request and deserialize response
        '''
        req = riak_pb.RpbSetBucketReq()
        req.bucket = str_to_bytes(bucket.name)
        self._add_bucket_type(req, bucket.bucket_type)

        self._encode_bucket_props(props, req)

        msg_code, resp = await self._request(
            messages.MSG_CODE_SET_BUCKET_REQ, req,
            messages.MSG_CODE_SET_BUCKET_RESP)
        return True

    def _add_bucket_type(self, req, bucket_type):
        if bucket_type and not bucket_type.is_default():
            req.type = bucket_type.name.encode()

    async def get_buckets(self, bucket_type=None):
        req = riak_pb.RpbListBucketsReq()
        if bucket_type:
            self._add_bucket_type(req, bucket_type)
        code, res = await self._request(messages.MSG_CODE_LIST_BUCKETS_REQ,
                                        req,
                                        messages.MSG_CODE_LIST_BUCKETS_RESP)
        return res.buckets

    async def get_keys(self, bucket):
        '''
        Lists all keys within a bucket.
        '''
        req = riak_pb.RpbListKeysReq()
        req.bucket = bucket.name.encode()
        keys = []
        self._add_bucket_type(req, bucket.bucket_type)
        for code, res in await self._stream(messages.MSG_CODE_LIST_KEYS_REQ,
                                            req,
                                            messages.MSG_CODE_LIST_KEYS_RESP):
            for key in res.keys:
                keys.append(key.decode())
        return keys

    def _decode_contents(self, contents, obj):
        '''
        Decodes the list of siblings from the protobuf representation
        into the object.

        :param contents: a list of RpbContent messages
        :type contents: list
        :param obj: a RiakObject
        :type obj: RiakObject
        :rtype RiakObject
        '''
        obj.siblings = [self._decode_content(c, RiakContent(obj))
                        for c in contents]
        # Invoke sibling-resolution logic
        if len(obj.siblings) > 1 and obj.resolver is not None:
            obj.resolver(obj)
        return obj

    def _decode_content(self, rpb_content, sibling):
        '''
        Decodes a single sibling from the protobuf representation into
        a RiakObject.

        :param rpb_content: a single RpbContent message
        :type rpb_content: riak_pb.RpbContent
        :param sibling: a RiakContent sibling container
        :type sibling: RiakContent
        :rtype: RiakContent
        '''

        if rpb_content.HasField("deleted") and rpb_content.deleted:
            sibling.exists = False
        else:
            sibling.exists = True
        if rpb_content.HasField("content_type"):
            sibling.content_type = rpb_content.content_type.decode()
        if rpb_content.HasField("charset"):
            sibling.charset = rpb_content.charset.decode()
        if rpb_content.HasField("content_encoding"):
            sibling.content_encoding = rpb_content.content_encoding.decode()
        if rpb_content.HasField("vtag"):
            sibling.etag = rpb_content.vtag.decode()

        sibling.links = [self._decode_link(link)
                         for link in rpb_content.links]
        if rpb_content.HasField("last_mod"):
            sibling.last_modified = float(rpb_content.last_mod)
            if rpb_content.HasField("last_mod_usecs"):
                sibling.last_modified += rpb_content.last_mod_usecs / 1000000.0

        sibling.usermeta = dict([(usermd.key.decode(), usermd.value.decode())
                                 for usermd in rpb_content.usermeta])
        sibling.indexes = set([(index.key,
                                decode_index_value(index.key, index.value))
                               for index in rpb_content.indexes])

        sibling.encoded_data = rpb_content.value

        return sibling

    def _decode_link(self, link):
        '''
        Decodes an RpbLink message into a tuple

        :param link: an RpbLink message
        :type link: riak_pb.RpbLink
        :rtype tuple
        '''

        if link.HasField("bucket"):
            bucket = link.bucket
        else:
            bucket = None
        if link.HasField("key"):
            key = link.key
        else:
            key = None
        if link.HasField("tag"):
            tag = link.tag
        else:
            tag = None

        return (bucket, key, tag)

    async def get(self, robj):
        '''
        Serialize get request and deserialize response
        '''
        bucket = robj.bucket

        req = riak_pb.RpbGetReq()
        req.bucket = bucket.name.encode()
        self._add_bucket_type(req, bucket.bucket_type)
        req.key = robj.key.encode()

        msg_code, resp = await self._request(messages.MSG_CODE_GET_REQ, req,
                                             messages.MSG_CODE_GET_RESP)
        if resp is not None:
            if resp.HasField('vclock'):
                robj.vclock = VClock(resp.vclock, 'binary')
            # We should do this even if there are no contents, i.e.
            # the object is tombstoned
            self._decode_contents(resp.content, robj)
        else:
            # "not found" returns an empty message,
            # so let's make sure to clear the siblings
            robj.siblings = []
        return robj

    async def put(self, robj, return_body=True):
        bucket = robj.bucket

        req = riak_pb.RpbPutReq()

        if return_body:
            req.return_body = 1

        req.bucket = str_to_bytes(bucket.name)
        self._add_bucket_type(req, bucket.bucket_type)

        if robj.key:
            req.key = str_to_bytes(robj.key)
        if robj.vclock:
            req.vclock = robj.vclock.encode('binary')

        self._encode_content(robj, req.content)

        msg_code, resp = await self._request(messages.MSG_CODE_PUT_REQ, req,
                                             messages.MSG_CODE_PUT_RESP)

        if resp is not None:
            if resp.HasField('key'):
                robj.key = bytes_to_str(resp.key)
            if resp.HasField('vclock'):
                robj.vclock = VClock(resp.vclock, 'binary')
            if resp.content:
                self._decode_contents(resp.content, robj)
        elif not robj.key:
            raise RiakError("missing response object")

        return robj

    async def delete(self, robj):
        req = riak_pb.RpbDelReq()

        use_vclocks = (hasattr(robj, 'vclock') and robj.vclock)
        if use_vclocks:
            req.vclock = robj.vclock.encode('binary')

        bucket = robj.bucket
        req.bucket = str_to_bytes(bucket.name)
        self._add_bucket_type(req, bucket.bucket_type)
        req.key = str_to_bytes(robj.key)

        msg_code, resp = await self._request(
            messages.MSG_CODE_DEL_REQ, req,
            messages.MSG_CODE_DEL_RESP)
        return self

    async def update_datatype(self, datatype, **options):

        if datatype.bucket.bucket_type.is_default():
            raise NotImplementedError("Datatypes cannot be used in the default"
                                      " bucket-type.")

        op = datatype.to_op()
        type_name = datatype.type_name
        if not op:
            raise ValueError("No operation to send on datatype {!r}".
                             format(datatype))

        req = riak_pb.DtUpdateReq()
        req.bucket = str_to_bytes(datatype.bucket.name)
        req.type = str_to_bytes(datatype.bucket.bucket_type.name)

        if datatype.key:
            req.key = str_to_bytes(datatype.key)
        if datatype._context:
            req.context = datatype._context

        self._encode_dt_options(req, options)
        self._encode_dt_op(type_name, req, op)

        msg_code, resp = await self._request(
            messages.MSG_CODE_DT_UPDATE_REQ, req,
            messages.MSG_CODE_DT_UPDATE_RESP)
        if resp.HasField('key'):
            datatype.key = resp.key[:]
        if resp.HasField('context'):
            datatype._context = resp.context[:]

        datatype._set_value(self._decode_dt_value(type_name, resp))

        return True



'''
Copyright 2013 Basho Technologies, Inc.
This file is provided to you under the Apache License,
Version 2.0 (the "License"); you may not use this file
except in compliance with the License.  You may obtain
a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
'''
from aioriak.error import RiakError


class RiakContent(object):
    '''
    The RiakContent holds the metadata and value of a single sibling
    within a RiakObject. RiakObjects that have more than one sibling
    are considered to be in conflict.
    '''
    def __init__(self, robject, data=None, encoded_data=None, charset=None,
                 content_type='application/json', content_encoding=None,
                 last_modified=None, etag=None, usermeta=None, links=None,
                 indexes=None, exists=False):
        self._robject = robject
        self._data = data
        self._encoded_data = encoded_data
        self.charset = charset
        self.content_type = content_type
        self.content_encoding = content_encoding
        self.last_modified = last_modified
        self.etag = etag
        self.usermeta = usermeta or {}
        self.links = links or []
        self.indexes = indexes or set()
        self.exists = exists

    def _get_data(self):
        if self._encoded_data is not None and self._data is None:
            self._data = self._deserialize(self._encoded_data)
            self._encoded_data = None
        return self._data

    def _set_data(self, value):
        self._encoded_data = None
        self._data = value

    data = property(_get_data, _set_data, doc='''
        The data stored in this object, as Python objects. For the raw
        data, use the `encoded_data` property. If unset, accessing
        this property will result in decoding the `encoded_data`
        property into Python values. The decoding is dependent on the
        `content_type` property and the bucket's registered decoders.
        :type mixed''')

    def _get_encoded_data(self):
        if self._data is not None and self._encoded_data is None:
            self._encoded_data = self._serialize(self._data)
            self._data = None
        return self._encoded_data

    def _set_encoded_data(self, value):
        self._data = None
        self._encoded_data = value

    encoded_data = property(_get_encoded_data, _set_encoded_data, doc='''
        The raw data stored in this object, essentially the encoded
        form of the `data` property. If unset, accessing this property
        will result in encoding the `data` property into a string. The
        encoding is dependent on the `content_type` property and the
        bucket's registered encoders.

        :type str''')

    def _serialize(self, value):
        encoder = self._robject.bucket.get_encoder(self.content_type)
        if encoder:
            return encoder(value)
        elif isinstance(value, str):
            return value.encode()
        else:
            raise TypeError('No encoder for non-string data '
                            'with content type "{0}"'.
                            format(self.content_type))

    def _deserialize(self, value):
        decoder = self._robject.bucket.get_decoder(self.content_type)
        if decoder:
            return decoder(value)
        else:
            raise TypeError('No decoder for content type "{0}"'.
                            format(self.content_type))

    def add_index(self, field, value):
        '''
        add_index(field, value)
        Tag this object with the specified field/value pair for
        indexing.

        :param field: The index field.
        :type field: string
        :param value: The index value.
        :type value: string or integer
        :rtype: :class:`RiakObject <riak.riak_object.RiakObject>`
        '''
        if field[-4:] not in ("_bin", "_int"):
            raise RiakError("Riak 2i fields must end with either '_bin'"
                            " or '_int'.")

        self.indexes.add((field, value))

        return self._robject

    def remove_index(self, field=None, value=None):
        '''
        remove_index(field=None, value=None)
        Remove the specified field/value pair as an index on this
        object.

        :param field: The index field.
        :type field: string
        :param value: The index value.
        :type value: string or integer
        :rtype: :class:`RiakObject <riak.riak_object.RiakObject>`
        '''
        if not field and not value:
            self.indexes.clear()
        elif field and not value:
            for index in [x for x in self.indexes if x[0] == field]:
                self.indexes.remove(index)
        elif field and value:
            self.indexes.remove((field, value))
        else:
            raise RiakError("Cannot pass value without a field"
                            " name while removing index")

        return self._robject

    remove_indexes = remove_index

    def set_index(self, field, value):
        '''
        Works like :meth:`add_index`, but ensures that there is only
        one index on given field. If other found, then removes it
        first.

        :param field: The index field.
        :type field: string
        :param value: The index value.
        :type value: string or integer
        :rtype: :class:`RiakObject <riak.riak_object.RiakObject>`
        '''
        to_rem = set((x for x in self.indexes if x[0] == field))
        self.indexes.difference_update(to_rem)
        return self.add_index(field, value)

    def add_link(self, obj, tag=None):
        '''
        Add a link to a RiakObject.

        :param obj: Either a RiakObject or 3 item link tuple consisting
            of (bucket, key, tag).
        :type obj: mixed
        :param tag: Optional link tag. Defaults to bucket name. It is ignored
            if ``obj`` is a 3 item link tuple.
        :type tag: string
        :rtype: :class:`RiakObject <riak.riak_object.RiakObject>`
        '''
        if isinstance(obj, tuple):
            newlink = obj
        else:
            newlink = (obj.bucket.name, obj.key, tag)

        self.links.append(newlink)
        return self._robject



from aioriak.datatypes.datatype import Datatype
from aioriak.datatypes import TYPES


class Flag(Datatype):
    '''
    A convergent datatype that represents a boolean value that can be
    enabled or disabled, and may only be embedded in :class:`Map`
    instances.
    '''

    type_name = 'flag'
    _type_error_msg = "Flags can only be booleans"

    def _post_init(self):
        self._op = None

    def _default_value(self):
        return False

    @Datatype.modified.getter
    def modified(self):
        '''
        Whether this flag has staged toggles.
        '''
        return self._op is not None

    def enable(self):
        '''
        Turns the flag on, effectively setting its value to ``True``.
        '''
        self._op = 'enable'

    def disable(self):
        '''
        Turns the flag off, effectively setting its value to ``False``.
        '''
        self._require_context()
        self._op = 'disable'

    def to_op(self):
        '''
        Extracts the mutation operation from the flag.

        :rtype: bool, None
        '''
        return self._op

    def _check_type(self, new_value):
        return isinstance(new_value, bool)


TYPES['flag'] = Flag



from setuptools import setup, find_packages
from commands import (docker_build, docker_start, docker_stop, setup_riak,
                      create_bucket_types, Test)

setup(
    name='aioriak',
    version='0.0.4',
    description='Async implementation of Riak DB python client',
    author='Makc Belousov',
    author_email='m.belousov@rambler-co.ru',
    url='https://github.com/rambler-digital-solutions/aioriak',
    keywords='riak asyncio client',
    packages=find_packages(exclude=('*.tests',)),
    include_package_data=True,
    zip_safe=False,
    license='MIT',
    install_requires=[
        'python3-riak-pb==2.1.0.6',
        'riak==2.3.0',
    ],
    tests_require=['nose==1.3.7',
                   'coverage==4.0.3'],
    cmdclass={
        'test': Test,
        'docker_build': docker_build,
        'docker_start': docker_start,
        'docker_stop': docker_stop,
        'setup_riak': setup_riak,
        'create_bucket_types': create_bucket_types,
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Database',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)



import logging
import json
from weakref import WeakValueDictionary
from .transport import create_transport
from .bucket import BucketType, Bucket
from riak.resolver import default_resolver
from riak.util import bytes_to_str, str_to_bytes
from aioriak.datatypes import TYPES


logger = logging.getLogger('aioriak.client')


def default_encoder(obj):
    '''
    Default encoder for JSON datatypes, which returns UTF-8 encoded
    json instead of the default bloated backslash u XXXX escaped ASCII strings.
    '''
    if isinstance(obj, bytes):
        return json.dumps(bytes_to_str(obj),
                          ensure_ascii=False).encode("utf-8")
    else:
        return json.dumps(obj, ensure_ascii=False).encode("utf-8")


def binary_json_encoder(obj):
    '''
    Default encoder for JSON datatypes, which returns UTF-8 encoded
    json instead of the default bloated backslash u XXXX escaped ASCII strings.
    '''
    if isinstance(obj, bytes):
        return json.dumps(bytes_to_str(obj),
                          ensure_ascii=False).encode("utf-8")
    else:
        return json.dumps(obj, ensure_ascii=False).encode("utf-8")


def binary_json_decoder(obj):
    '''
    Default decoder from JSON datatypes.
    '''
    return json.loads(bytes_to_str(obj))


def binary_encoder_decoder(obj):
    '''
    Assumes value is already in binary format, so passes unchanged.
    '''
    return obj


class RiakClient:
    '''
    The ``RiakClient`` object holds information necessary to connect
    to Riak. Requests can be made to Riak directly through the client
    or by using the methods on related objects.
    '''
    def __init__(self, host='localhost', port=8087, loop=None):
        self._host = host
        self._port = port
        self._loop = loop
        self._bucket_types = WeakValueDictionary()
        self._buckets = WeakValueDictionary()
        self._resolver = None
        self._decoders = {'application/json': binary_json_decoder,
                          'text/json': binary_json_decoder,
                          'text/plain': bytes_to_str,
                          'binary/octet-stream': binary_encoder_decoder}
        self._encoders = {'application/json': binary_json_encoder,
                          'text/json': binary_json_encoder,
                          'text/plain': str_to_bytes,
                          'binary/octet-stream': binary_encoder_decoder}

    def get_decoder(self, content_type):
        '''
        Get the decoding function for the provided content type.

        :param content_type: the requested media type
        :type content_type: str
        :rtype: function
        '''
        return self._decoders.get(content_type)

    def get_encoder(self, content_type):
        '''
        Get the encoding function for the provided content type.

        :param content_type: the requested media type
        :type content_type: str
        :rtype: function
        '''
        return self._encoders.get(content_type)

    def set_encoder(self, content_type, encoder):
        '''
        Set the encoding function for the provided content type.

        :param content_type: the requested media type
        :type content_type: str
        :param encoder: an encoding function, takes a single object
            argument and returns encoded data
        :type encoder: function
        '''
        self._encoders[content_type] = encoder

    def set_decoder(self, content_type, decoder):
        '''
        Set the decoding function for the provided content type.

        :param content_type: the requested media type
        :type content_type: str
        :param decoder: a decoding function, takes encoded data and
            returns a Python type
        :type decoder: function
        '''
        self._decoders[content_type] = decoder

    def _get_resolver(self):
        return self._resolver or default_resolver

    def _set_resolver(self, value):
        if value is None or callable(value):
            self._resolver = value
        else:
            raise TypeError('resolver is not a function')

    resolver = property(_get_resolver, _set_resolver,
                        doc='''The sibling-resolution function for this client.
                        Defaults to :func:`riak.resolver.default_resolver`.''')

    def close(self):
        self._transport.close()

    async def _create_transport(self):
        self._transport = await create_transport(
            self._host, self._port, self._loop)

    @classmethod
    async def create(cls, host='localhost', port=8087, loop=None):
        '''
        Return initialized instance of RiakClient since
        RiakClient.__init__() can't be async.

        :Example:

        .. code-block:: python

            import asyncio
            from aioriak import RiakClient
            loop = asyncio.get_event_loop()
            async def go():
                client = await RiakClient.create('localhost', 8087, loop)
            loop.run_until_complete(go())

        :param host: Hostname or ip address of Riak instance
        :type host: str
        :param port: Port of riak instance
        :type port: int
        :param loop: asyncio event loop
        :rtype: :class:`~aioriak.client.RiakClient`
        '''
        client = cls(host, port, loop)
        await client._create_transport()
        return client

    async def fetch_datatype(self, bucket, key):
        '''
        Fetches the value of a Riak Datatype.

        :param bucket: the bucket of the datatype, which must belong to a
            :class:`~aioriak.bucket.BucketType`
        :type bucket: :class:`~aioriak.bucket.Bucket`
        :param key: the key of the datatype
        :type key: string
        :rtype: :class:`~aioriak.datatypes.Datatype`
        '''
        dtype, value, context = await self._fetch_datatype(bucket, key)

        return TYPES[dtype](bucket=bucket, key=key, value=value,
                            context=context)

    async def _fetch_datatype(self, bucket, key):
        '''
        _fetch_datatype(bucket, key)

        Fetches the value of a Riak Datatype as raw data. This is used
        internally to update already reified Datatype objects. Use the
        public version to fetch a reified type.

        :param bucket: the bucket of the datatype, which must belong to a
            :class:`~aioriak.BucketType`
        :type bucket: RiakBucket
        :param key: the key of the datatype
        :type key: string, None
        :rtype: tuple of type, value and context
        '''
        return await self._transport.fetch_datatype(bucket, key)

    async def ping(self):
        '''
        Check if the Riak server for this ``RiakClient`` instance is alive.

        :rtype: boolean
        '''
        return await self._transport.ping()

    is_alive = ping

    async def get_client_id(self):
        '''
        Get client ID for this RiakClient instance

        :rtype: bytes
        '''
        return await self._transport.get_client_id()

    async def set_client_id(self, id):
        '''
        Set Client ID for this RiakClient instance'
        '''
        return await self._transport.set_client_id(id)

    async def get_buckets(self, bucket_type=None):
        '''
        Get the list of buckets as :class:`Bucket
        <aioriak.bucket.Bucket>` instances.

        .. warning:: Do not use this in production, as it requires
            traversing through all keys stored in a cluster.

        :param bucket_type: the optional containing bucket type
        :type bucket_type: :class:`~aioriak.bucket.BucketType`
        :rtype: list of :class:`Bucket <aioriak.bucket.Bucket>`
            instances
        '''
        if bucket_type:
            maker = bucket_type.bucket
        else:
            maker = self.bucket

        return [maker(name.decode())
                for name in await self._transport.get_buckets(
                    bucket_type=bucket_type)]

    def bucket_type(self, name):
        '''
        Gets the bucket-type by the specified name. Bucket-types do
        not always exist (unlike buckets), but this will always return
        a :class:`BucketType <aioriak.bucket.BucketType>` object.

        :param name: the bucket name
        :type name: str
        :rtype: :class:`BucketType <aioriak.bucket.BucketType>`
        '''
        if not isinstance(name, str):
            raise TypeError('Bucket name must be a string')

        if name in self._bucket_types:
            return self._bucket_types[name]
        else:
            btype = BucketType(self, name)
            self._bucket_types[name] = btype
            return btype

    def bucket(self, name, bucket_type='default'):
        '''
        Get the bucket by the specified name. Since buckets always exist,
        this will always return a
        :class:`Bucket <aioriak.bucket.Bucket>`.
        If you are using a bucket that is contained in a bucket type, it is
        preferable to access it from the bucket type object::

            # Preferred:
            client.bucket_type("foo").bucket("bar")
            # Equivalent, but not preferred:
            client.bucket("bar", bucket_type="foo")

        :param name: the bucket name
        :type name: str
        :param bucket_type: the parent bucket-type
        :type bucket_type: :class:`BucketType <aioriak.bucket.BucketType>`
            or str
        :rtype: :class:`Bucket <aioriak.bucket.Bucket>`
        '''
        if not isinstance(name, str):
            raise TypeError('Bucket name must be a string')

        if isinstance(bucket_type, str):
            bucket_type = self.bucket_type(bucket_type)
        elif not isinstance(bucket_type, BucketType):
            raise TypeError('bucket_type must be a string '
                            'or aioriak.bucket.BucketType')

        return self._buckets.setdefault((bucket_type, name),
                                        Bucket(self, name, bucket_type))

    async def get_bucket_type_props(self, bucket_type):
        '''
        Fetches properties for the given bucket-type.

        :param bucket_type: the bucket-type whose properties will be fetched
        :type bucket_type: BucketType
        :rtype: dict
        '''
        return await self._transport.get_bucket_type_props(bucket_type)

    async def set_bucket_type_props(self, bucket_type, props):
        '''
        Sets properties for the given bucket-type.

        :param bucket_type: the bucket-type whose properties will be set
        :type bucket_type: BucketType
        :param props: the properties to set
        :type props: dict
        '''
        return await self._transport.set_bucket_type_props(bucket_type, props)

    async def get_bucket_props(self, bucket):
        '''
        Fetches bucket properties for the given bucket.

        :param bucket: the bucket whose properties will be fetched
        :type bucket: Bucket
        :rtype: dict
        '''
        return await self._transport.get_bucket_props(bucket)

    async def set_bucket_props(self, bucket, props):
        '''
        Sets bucket properties for the given bucket.

        :param bucket: the bucket whose properties will be set
        :type bucket: Bucket
        :param props: the properties to set
        :type props: dict
        '''
        return await self._transport.set_bucket_props(bucket, props)

    async def get_keys(self, bucket):
        '''
        Lists all keys in a bucket.

        .. warning:: Do not use this in production, as it requires
           traversing through all keys stored in a cluster.

        :param bucket: the bucket whose keys are fetched
        :type bucket: Bucket
        :rtype: list
        '''
        return await self._transport.get_keys(bucket)

    async def get(self, robj):
        '''
        Fetches the contents of a Riak object.

        :param robj: the object to fetch
        :type robj: RiakObject
        '''
        if not isinstance(robj.key, str):
            raise TypeError(
                'key must be a string, instead got {0}'.format(repr(robj.key)))

        return await self._transport.get(robj)

    async def put(self, robj, return_body):
        '''
        Stores an object in the Riak cluster.

        :param return_body: whether to return the resulting object
            after the write
        :type return_body: boolean
        :param robj: the object to store
        :type robj: RiakObject
        '''
        return await self._transport.put(robj, return_body=return_body)

    async def delete(self, robj):
        '''
        Deletes an object from Riak.

        :param robj: the object to delete
        :type robj: RiakObject
        '''
        return await self._transport.delete(robj)

    async def update_datatype(self, datatype, **params):
        '''
        Sends an update to a Riak Datatype to the server.

        :param datatype: the datatype with pending updates
        :type datatype: :class:`~aioriak.datatypes.Datatype`
        :rtype: tuple of datatype, opaque value and opaque context
        '''

        return await self._transport.update_datatype(datatype, **params)



from .types import TYPES
from .datatype import Datatype
from .counter import Counter
from .map import Map
from .register import Register
from .flag import Flag
from .set import Set


__all__ = ['Datatype', 'Map', 'TYPES', 'Counter', 'Register', 'Flag', 'Set']



from collections import Sized
from aioriak.datatypes.datatype import Datatype
from aioriak.datatypes import TYPES


class Register(Sized, Datatype):
    '''
    A convergent datatype that represents an opaque string that is set
    with last-write-wins semantics, and may only be embedded in
    :class:`~aioriak.datatypes.Map` instances.
    '''

    type_name = 'register'
    _type_error_msg = 'Registers can only be strings'

    def _post_init(self):
        self._new_value = None

    def _default_value(self):
        return ''

    @Datatype.value.getter
    def value(self):
        '''
        Returns a copy of the original value of the register.

        :rtype: str
        '''
        return self._value[:]

    @Datatype.modified.getter
    def modified(self):
        '''
        Whether this register has staged assignment.
        '''
        return self._new_value is not None

    def to_op(self):
        '''
        Extracts the mutation operation from the register.

        :rtype: str, None
        '''
        if self._new_value is not None:
            return ('assign', self._new_value)

    def assign(self, new_value):
        '''
        Assigns a new value to the register.

        :param new_value: the new value for the register
        :type new_value: str
        '''
        self._raise_if_badtype(new_value)
        self._new_value = new_value

    def __len__(self):
        return len(self.value)

    def _check_type(self, new_value):
        return isinstance(new_value, str)


TYPES['register'] = Register
